#define ATTENUATION
//#define HQ_ATTENUATION

varying vec2 texCoord;

varying vec4 AmbientSum;
varying vec4 DiffuseSum;
varying vec4 SpecularSum;

varying vec3 vPosition;
varying vec3 vViewDir;
varying vec4 vLightDir;


#ifdef DIFFUSEMAP
  uniform sampler2D m_DiffuseMap;
#endif

#ifdef DIFFUSEMAP_BACK
  uniform sampler2D m_DiffuseMapBack;
  varying vec2 backgroundTexCoord;
#endif

#ifdef PARALLAXMAP
  uniform sampler2D m_ParallaxMap;
#endif
  
#ifdef NORMALMAP
  uniform sampler2D m_NormalMap;
#else
  varying vec3 vNormal;
#endif

#ifdef ALPHAMAP
  uniform sampler2D m_AlphaMap;
#endif

#ifdef DECAL_COLOR
uniform vec4 m_DecalColor;
#endif

#ifdef ENVMAP
uniform sampler2D m_EnvMap;
uniform float m_DecalMetal;
uniform float m_BackMetal;

varying vec2 reflectTexCoord;
#endif


uniform vec4 g_AmbientLightColor;
uniform float m_Shininess;

#ifdef HQ_ATTENUATION
uniform vec4 g_LightPosition;
varying vec3 lightVec;
#endif


float tangDot(in vec3 v1, in vec3 v2){
    float d = dot(v1,v2);
    #ifdef V_TANGENT
        d = 1.0 - d*d;
        return step(0.0, d) * sqrt(d);
    #else
        return d;
    #endif
}

float lightComputeDiffuse(in vec3 norm, in vec3 lightdir, in vec3 viewdir){
    #ifdef MINNAERT
        float NdotL = max(0.0, dot(norm, lightdir));
        float NdotV = max(0.0, dot(norm, viewdir));
        return NdotL * pow(max(NdotL * NdotV, 0.1), -1.0) * 0.5;
    #else
        return max(0.0, dot(norm, lightdir));
    #endif
}

float lightComputeSpecular(in vec3 norm, in vec3 viewdir, in vec3 lightdir, in float shiny){
    
    #if defined(WARDISO)
        // Isotropic Ward
        vec3 halfVec = normalize(viewdir + lightdir);
        float NdotH  = max(0.001, tangDot(norm, halfVec));
        float NdotV  = max(0.001, tangDot(norm, viewdir));
        float NdotL  = max(0.001, tangDot(norm, lightdir));
        float a      = tan(acos(NdotH));
        float p      = max(shiny/128.0, 0.001);
        return NdotL * (1.0 / (4.0*3.14159265*p*p)) * (exp(-(a*a)/(p*p)) / (sqrt(NdotV * NdotL)));
    #else
       // Standard Phong
       vec3 R = reflect(-lightdir, norm);
       return pow(max(tangDot(R, viewdir), 0.0), shiny);
    #endif
}

vec2 computeLighting(in vec3 wvPos, in vec3 wvNorm, in vec3 wvViewDir, in vec3 wvLightDir){
   float diffuseFactor = lightComputeDiffuse(wvNorm, wvLightDir, wvViewDir);
   float specularFactor = lightComputeSpecular(wvNorm, wvViewDir, wvLightDir, m_Shininess);
   specularFactor *= step(1.0, m_Shininess);

   #ifdef HQ_ATTENUATION
    float att = clamp(1.0 - g_LightPosition.w * length(lightVec), 0.0, 1.0);
   #else
    float att = vLightDir.w;
   #endif

   return vec2(diffuseFactor, specularFactor) * vec2(att);
}


void main(){
    vec2 newTexCoord;
    #if defined(PARALLAXMAP) || defined(NORMALMAP_PARALLAX)
       float h;
       #ifdef PARALLAXMAP
          h = texture2D(m_ParallaxMap, texCoord).r;
       #else
          h = texture2D(m_NormalMap, texCoord).a;
       #endif
       float heightScale = 0.05;
       float heightBias = heightScale * -0.5;
       vec3 normView = normalize(vViewDir);
       h = (h * heightScale + heightBias) * normView.z;
       newTexCoord = texCoord + (h * -normView.xy);
    #else
       newTexCoord = texCoord;
    #endif
    

    // ***********************
    // Read from textures
    // ***********************
    #if defined(NORMALMAP)
      vec4 normalHeight = texture2D(m_NormalMap, newTexCoord);
      vec3 normal = (normalHeight.xyz * vec3(2.0) - vec3(1.0));
      #ifdef LATC
        normal.z = sqrt(1.0 - (normal.x * normal.x) - (normal.y * normal.y));
      #endif
      normal.y = -normal.y;
    #else
      vec3 normal = vNormal;
      #if !defined(LOW_QUALITY) && !defined(V_TANGENT)
         normal = normalize(normal);
      #endif
    #endif

      
      //color.rgb = color.rgb*(1-alp) + m_DecalColor.rgb*(alp);
      //color.a = color.a*(1-alp) + m_DecalColor.a*(alp);

	vec4 diffuseColor;

    #ifdef DIFFUSEMAP
      float alp = texture2D(m_DiffuseMap, newTexCoord).a;
      diffuseColor.rgb = m_DecalColor.rgb*alp;
      diffuseColor.a = alp;
    #else
      diffuseColor = vec4(1.0);
    #endif
    
    
    //TODO: check how alfa plays with transparency
    //TODO: make separate 'metal' factor for decal and for background
    #ifdef ENVMAP
    	vec4 env = texture2D( m_EnvMap, reflectTexCoord.st);
    	float useMetal = (m_DecalMetal*diffuseColor.a) + (m_BackMetal*(1-diffuseColor.a));
    	diffuseColor.rgb = diffuseColor.rgb * (1-useMetal)  + env.rgb * (useMetal);
    	diffuseColor.a = max(diffuseColor.a,useMetal);
    #endif

    float alpha = max(DiffuseSum.a, diffuseColor.a);
    #ifdef ALPHAMAP
       alpha = alpha * texture2D(m_AlphaMap, newTexCoord).r;
    #endif

       
       
    vec4 lightDir = vLightDir;
    lightDir.xyz = normalize(lightDir.xyz);
    vec2 light = computeLighting(vPosition, normal, vViewDir.xyz, lightDir.xyz); 

	vec4 finalColor;

    #ifdef DIFFUSEMAP_BACK
    	vec4 backColor = texture2D(m_DiffuseMapBack, backgroundTexCoord);
    	
    	finalColor.rgb = (g_AmbientLightColor.rgb + vec3(1.0)* light.x)  * diffuseColor.rgb * diffuseColor.a +
       				  (AmbientSum.rgb + DiffuseSum.rgb * light.x) * (1-diffuseColor.a)*backColor.rgb +
                      SpecularSum.rgb * light.y;
    
    #else
    	finalColor.rgb = (g_AmbientLightColor.rgb + vec3(1.0)* light.x)  * diffuseColor.rgb * diffuseColor.a +
       				  (AmbientSum.rgb + DiffuseSum.rgb * light.x) * (1-diffuseColor.a) +
                      SpecularSum.rgb * light.y;
    
    #endif
    

	finalColor.a = alpha;
       
    gl_FragColor = finalColor;
}
